package com.chelvc.framework.database.interceptor;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;

import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.chelvc.framework.common.exception.ParameterInvalidException;
import com.chelvc.framework.common.util.ObjectUtils;
import com.chelvc.framework.database.context.DatabaseContextHolder;
import com.chelvc.framework.database.context.UniqueContext;
import com.chelvc.framework.database.entity.Entity;
import lombok.extern.slf4j.Slf4j;
import org.apache.ibatis.cache.CacheKey;
import org.apache.ibatis.executor.Executor;
import org.apache.ibatis.mapping.BoundSql;
import org.apache.ibatis.mapping.MappedStatement;
import org.apache.ibatis.mapping.SqlCommandType;
import org.apache.ibatis.plugin.Interceptor;
import org.apache.ibatis.plugin.Intercepts;
import org.apache.ibatis.plugin.Invocation;
import org.apache.ibatis.plugin.Plugin;
import org.apache.ibatis.plugin.Signature;
import org.apache.ibatis.session.ResultHandler;
import org.apache.ibatis.session.RowBounds;
import org.springframework.stereotype.Component;

/**
 * 基础属性自动更新维护拦截器
 *
 * @author Woody
 * @date 2024/1/30
 */
@Slf4j
@Component
@Intercepts({
        @Signature(type = Executor.class, method = "update", args = {MappedStatement.class, Object.class}),
        @Signature(type = Executor.class, method = "query", args = {MappedStatement.class, Object.class,
                RowBounds.class, ResultHandler.class}),
        @Signature(type = Executor.class, method = "query", args = {MappedStatement.class, Object.class,
                RowBounds.class, ResultHandler.class, CacheKey.class, BoundSql.class})
})
public class PropertyUpdateInterceptor implements Interceptor {
    /**
     * 属性更新拦截处理
     *
     * @param invocation 调用信息
     * @return 调用结果
     * @throws Throwable 调用异常
     */
    @SuppressWarnings("unchecked")
    private Object processing(Invocation invocation) throws Throwable {
        SqlCommandType operation = ((MappedStatement) invocation.getArgs()[0]).getSqlCommandType();
        if (operation == SqlCommandType.SELECT) {
            // 加密数据解码
            Object result = invocation.proceed();
            if (ObjectUtils.notBlank(result) && DatabaseContextHolder.isCodecEnabled()) {
                try {
                    DatabaseContextHolder.decoder(cipher -> ObjectUtils.stream(result).forEach(
                            entity -> DatabaseContextHolder.decode(cipher, entity)
                    ));
                } catch (Exception e) {
                    log.warn("Confidential value decode failed: {}", e.getMessage());
                }
            }
            return result;
        } else if (operation == SqlCommandType.INSERT || operation == SqlCommandType.UPDATE) {
            Object entity = DatabaseContextHolder.getOriginalEntity(invocation.getArgs()[1]);
            if (entity instanceof Entity) {
                // 初始化默认属性值
                DatabaseContextHolder.initializeEntityDefaultValues((Entity<?>) entity);
            }

            // 设置加密属性值
            Class<Object> clazz = (Class<Object>) entity.getClass();
            boolean encoding = DatabaseContextHolder.isCodecEnabled() && DatabaseContextHolder.isConfidential(clazz);
            Object original = encoding ? ObjectUtils.copying(entity, clazz) : entity;
            if (encoding) {
                DatabaseContextHolder.encode(entity);
            }

            // 唯一属性值校验
            // 根据唯一属性值查询现有记录
            List<UniqueContext> uniques = DatabaseContextHolder.getUniqueContexts(clazz);
            if (ObjectUtils.isEmpty(uniques)) {
                return invocation.proceed();
            }
            LambdaQueryWrapper<Object> query = DatabaseContextHolder.initializeUniqueQuery(clazz, entity, uniques);
            List<Object> objects = query.isEmptyOfWhere() ? Collections.emptyList() :
                    DatabaseContextHolder.lookupEntityService(clazz).list(query);
            if (ObjectUtils.isEmpty(objects)) {
                return invocation.proceed();
            }

            // 如果是修改，则排除主键相同的数据
            if (operation == SqlCommandType.UPDATE) {
                Object id = DatabaseContextHolder.getEntityIdentity(entity);
                objects.removeIf(object -> Objects.equals(id, DatabaseContextHolder.getEntityIdentity(object)));
            }

            // 比较唯一属性值，找出已存在唯一属性名称
            Map<String, String> errors = uniques.stream().filter(context -> objects.stream().anyMatch(
                    object -> context.getGetters().stream().allMatch(getter -> {
                        Object value = getter.apply(object);
                        if (encoding) {
                            return Objects.equals(value, getter.apply(original))
                                    || Objects.equals(value, getter.apply(entity));
                        }
                        return Objects.equals(value, getter.apply(original));
                    })
            )).collect(Collectors.toMap(
                    context -> context.getField().getName(), context -> context.getUnique().message()
            ));
            if (ObjectUtils.notEmpty(errors)) {
                throw new ParameterInvalidException(errors);
            }
        }
        return invocation.proceed();
    }

    @Override
    public Object intercept(Invocation invocation) throws Throwable {
        if (invocation.getTarget() instanceof Executor) {
            return this.processing(invocation);
        }
        return invocation.proceed();
    }

    @Override
    public Object plugin(Object target) {
        return target instanceof Executor ? Plugin.wrap(target, this) : target;
    }
}
